import numpy as np
import open3d as o3
import scipy.spatial as ss
from . import sdp


def to_pointcloud(arr):
    pc = o3.PointCloud()
    pc.points = o3.Vector3dVector(arr.T)
    return pc


def sdrsac(m, b, max_itr=10000,
           n_sample=9,
           k=4,
           d_diff_thresh=1.0e-4,
           pair_dist_thresh=1.0e-2,
           n_pair_thres=5,
           corr_count_eps=0.015,
           icp_thres=0.02,
           callbacks=[]):
    btree = ss.KDTree(b.T)
    max_inls = 0
    ps = 0.99
    best_rot = np.identity(3)
    best_t = np.zeros(3)
    iter = 0
    t_max = 1e10
    stop = False
    while iter < max_itr and not stop:
        idx_m = np.random.choice(m.shape[1], n_sample, False)
        smp_m = m[:, idx_m]
        b_to_sample = b
        scount = 0
        while b_to_sample.shape[1] > n_sample * 2.0 and scount < 2:
            scount += 1
            idx_b = np.random.choice(b_to_sample.shape[1], n_sample, False)
            smp_b = b_to_sample[:, idx_b]
            b_to_sample = np.delete(b_to_sample, idx_b, axis=1)

            rot, t, _, corr_b = sdp.sdp_reg(smp_m, smp_b,
                                            k, d_diff_thresh,
                                            pair_dist_thresh,
                                            n_pair_thres)

            if not corr_b is None and len(corr_b) > 0:
                tm = (np.dot(rot, m).transpose() + t).transpose()
                reg_p2p = o3.registration.registration_icp(to_pointcloud(tm), to_pointcloud(smp_b),
                                                           icp_thres, np.identity(4),
                                                           o3.registration.TransformationEstimationPointToPoint())
                rot_icp = reg_p2p.transformation[:3, :3]
                t_icp = reg_p2p.transformation[:3, 3]
                tm_icp = (np.dot(rot_icp, tm).transpose() + t_icp).transpose()
                inls_icp = sdp.count_correspondence(tm_icp, btree, corr_count_eps)
                if inls_icp > max_inls:
                    max_inls = inls_icp
                    best_rot = np.dot(rot_icp, rot)
                    best_t = np.dot(rot_icp, t) + t_icp
                    p_i = max_inls / m.shape[1]
                    t_max = np.log(1.0 - ps) / np.log(1.0 - p_i**k)
                    for c in callbacks:
                        c(best_rot, best_t)
            iter += iter
            if iter >= t_max and iter >= 5:
                stop = True
    return best_rot, best_t


from . import _utils


def generate_weight_matrix(m, b, d_diff_thresh, pair_dist_thresh):
    return _utils.generate_weight_matrix(m, b, d_diff_thresh, pair_dist_thresh)


def generate_sdp_constraint_map(weight, n, k):
    return _utils.generate_sdp_constraint_map(weight, n, k)


import copy
import numpy as np
import open3d as o3


class Open3dVisualizerCallback(object):
    """Display the 3D registration result of each iteration.

    Args:
        source (numpy.ndarray): Source point cloud data.
        target (numpy.ndarray): Target point cloud data.
        save (bool, optional): If this flag is True,
            each iteration image is saved in a sequential number.
        keep_window (bool, optional): If this flag is True,
            the drawing window blocks after registration is finished.
    """
    def __init__(self, source, target, save=False,
                 keep_window=True):
        self._vis = o3.Visualizer()
        self._vis.create_window()
        self._source = source
        self._target = target
        self._result = copy.deepcopy(self._source)
        self._save = save
        self._keep_window = keep_window
        self._source.paint_uniform_color([1, 0, 0])
        self._target.paint_uniform_color([0, 1, 0])
        self._result.paint_uniform_color([0, 0, 1])
        self._vis.add_geometry(self._source)
        self._vis.add_geometry(self._target)
        self._vis.add_geometry(self._result)
        self._cnt = 0

    def __del__(self):
        if self._keep_window:
            self._vis.run()
        self._vis.destroy_window()

    def __call__(self, rot, t):
        self._result.points = copy.deepcopy(self._source.points)
        trans = np.zeros((4, 4))
        trans[:3, :3] = rot
        trans[:3, 3] = t
        self._result.transform(trans)
        self._vis.update_geometry()
        self._vis.poll_events()
        self._vis.update_renderer()
        if self._save:
            self._vis.capture_screen_image("image_%04d.jpg" % self._cnt)
        self._cnt += 1



import numpy as np
import picos
from cvxopt import matrix, spmatrix
from . import linear_projection
from . import utils
from . import _kabsch


def to_spmatrix(sp):
    coo = sp.astype(np.double).tocoo()
    return spmatrix(coo.data.tolist(), coo.row.tolist(), coo.col.tolist(), size=sp.shape)


def solve_sdp(w, k):
    nsquare = w.shape[0]
    n = int(np.sqrt(nsquare))
    ww = np.c_[w, np.zeros((nsquare, 1))]
    ww = np.r_[ww, np.zeros((1, nsquare + 1))]
    maps, mapeq, maps_b, mapeq_b = utils.generate_sdp_constraint_map(ww, n, k)

    sdp = picos.Problem()
    y = sdp.add_variable('y', (nsquare + 1, nsquare + 1), 'symmetric')
    sdp.add_constraint(y >> 0)
    sdp.add_constraint(y[nsquare, nsquare] == 1.0)
    sdp.add_constraint(y[nsquare, :] == y[:, nsquare].T)
    for i, m in enumerate(maps):
        sdp.add_constraint(to_spmatrix(m) | y <= maps_b.astype(np.double)[i, 0])
    sdp.add_constraint(to_spmatrix(mapeq) | y == matrix(mapeq_b))
    sdp.add_constraint(picos.trace(y) == k + 1)
    sdp.set_objective('max', matrix(ww) | y)
    sdp.set_option('tol', 0.1)
    sdp.set_option('verbose', 0)
    # print(sdp)
    sdp.solve()

    y = y.value
    t = np.array(y[-1, :-1])
    x1 = t.reshape((n, n))
    x = linear_projection.linear_projection(x1.ravel(order='F'))

    idx, = np.where(x.ravel(order='F')==1.0)
    score = x1.ravel(order='F')[idx]
    sidx = np.argsort(score)
    idx = np.delete(idx, sidx[(n - k):])
    x.ravel(order='F')[idx] = 0.0
    return x


def count_correspondence(m, tree, eps):
    dist, idx = tree.query(m.T)
    in_idx, = np.where(dist <= eps)
    return len(in_idx)


def get_correspondences(x):
    idxs = np.argmax(x, axis=1)
    y = x[np.arange(x.shape[0]), idxs]
    corr_m = np.where(y > 0)[0]
    corr_b = idxs[corr_m]
    return corr_m, corr_b


def sdp_reg(m, b, k, d_diff_thresh, pair_dist_thresh, n_pair_thres):
    w, n_accepted_pairs = utils.generate_weight_matrix(m, b,
                                                       d_diff_thresh,
                                                       pair_dist_thresh)
    if n_accepted_pairs <= n_pair_thres:
        return None, None, None, None

    x = solve_sdp(w, k)
    corr_m, corr_b = get_correspondences(x)
    sm = m[:, corr_m]
    sb = b[:, corr_b]
    rot, t = _kabsch.kabsch(sm, sb)
    return rot, t, corr_m, corr_b


import numpy as np
from cvxopt import matrix, solvers


def linear_projection(xp):
    f = xp.copy()
    f[f < 0.0] = 0.0
    f = -f
    nn = xp.shape[0]
    n = int(np.round(np.sqrt(nn)))

    aa = np.zeros((4 * n, nn))
    bb = np.tile(np.array([1.0, -1.0]), 2 * n)
    for i in range(n):
        ta = np.zeros((n, n))
        ta[i, :] = 1.0
        tar = ta.ravel(order='F')
        aa[4 * i, :] = tar
        aa[4 * i + 1, :] = -tar

        ta = np.zeros((n, n))
        ta[:, i] = 1.0
        tar = ta.ravel(order='F')
        aa[4 * i + 2, :] = tar
        aa[4 * i + 3, :] = -tar

    aa = np.r_[aa, -np.identity(nn)]
    bb = np.r_[bb, np.zeros(nn)]
    sol = solvers.lp(matrix(f), matrix(aa), matrix(bb))
    return np.array(sol['x']).reshape((n, n))


import os
import re
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
import sys
import setuptools

__version__ = '0.0.1'

def find_eigen(hint=[]):
    """
    Find the location of the Eigen 3 include directory. This will return
    ``None`` on failure.
    """
    # List the standard locations including a user supplied hint.
    search_dirs = hint + [
        "/usr/local/include/eigen3",
        "/usr/local/homebrew/include/eigen3",
        "/opt/local/var/macports/software/eigen3",
        "/opt/local/include/eigen3",
        "/usr/include/eigen3",
        "/usr/include/local",
        "/usr/include",
    ]

    # Loop over search paths and check for the existence of the Eigen/Dense
    # header.
    for d in search_dirs:
        path = os.path.join(d, "Eigen", "Dense")
        if os.path.exists(path):
            # Determine the version.
            vf = os.path.join(d, "Eigen", "src", "Core", "util", "Macros.h")
            if not os.path.exists(vf):
                continue
            src = open(vf, "r").read()
            v1 = re.findall("#define EIGEN_WORLD_VERSION (.+)", src)
            v2 = re.findall("#define EIGEN_MAJOR_VERSION (.+)", src)
            v3 = re.findall("#define EIGEN_MINOR_VERSION (.+)", src)
            if not len(v1) or not len(v2) or not len(v3):
                continue
            v = "{0}.{1}.{2}".format(v1[0], v2[0], v3[0])
            print("Found Eigen version {0} in: {1}".format(v, d))
            return d
    return None

class get_pybind_include(object):
    """Helper class to determine the pybind11 include path
    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked. """

    def __init__(self, user=False):
        try:
            import pybind11
        except ImportError:
            if subprocess.call([sys.executable, '-m', 'pip', 'install', 'pybind11']):
                raise RuntimeError('pybind11 install failed.')
        self.user = user

    def __str__(self):
        import pybind11
        return pybind11.get_include(self.user)


ext_modules = [
    Extension(
        'sdrsac._utils',
        ['sdrsac/cc/utils_py.cc', 'sdrsac/cc/utils.cc'],
        include_dirs=[
            # Path to pybind11 headers
            './sdrsac',
            get_pybind_include(),
            get_pybind_include(user=True),
            find_eigen(['third_party/eigen'])
        ],
        language='c++'
    ),
    Extension(
        'sdrsac._kabsch',
        ['sdrsac/cc/kabsch_py.cc'],
        include_dirs=[
            # Path to pybind11 headers
            './sdrsac',
            get_pybind_include(),
            get_pybind_include(user=True),
            find_eigen(['third_party/eigen'])
        ],
        language='c++'
    ),
]


# As of Python 3.6, CCompiler has a `has_flag` method.
# cf http://bugs.python.org/issue26689
def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    import tempfile
    with tempfile.NamedTemporaryFile('w', suffix='.cpp') as f:
        f.write('int main (int argc, char **argv) { return 0; }')
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except setuptools.distutils.errors.CompileError:
            return False
    return True


def cpp_flag(compiler):
    """Return the -std=c++[11/14] compiler flag.
    The c++14 is prefered over c++11 (when it is available).
    """
    if has_flag(compiler, '-std=c++14'):
        return '-std=c++14'
    elif has_flag(compiler, '-std=c++11'):
        return '-std=c++11'
    else:
        raise RuntimeError('Unsupported compiler -- at least C++11 support '
                           'is needed!')


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""
    c_opts = {
        'msvc': ['/EHsc'],
        'unix': [],
    }

    if sys.platform == 'darwin':
        c_opts['unix'] += ['-stdlib=libc++', '-mmacosx-version-min=10.7']

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        if ct == 'unix':
            opts.append('-DVERSION_INFO="%s"' % self.distribution.get_version())
            opts.append(cpp_flag(self.compiler))
            if has_flag(self.compiler, '-fvisibility=hidden'):
                opts.append('-fvisibility=hidden')
        elif ct == 'msvc':
            opts.append('/DVERSION_INFO=\\"%s\\"' % self.distribution.get_version())
        for ext in self.extensions:
            ext.extra_compile_args = opts
        build_ext.build_extensions(self)

setup(
    name='sdrsac',
    version=__version__,
    packages=['sdrsac'],
    author='neka-nat',
    author_email='nekanat.stock@gmail.com',
    url='https://github.com/neka-nat/sdrsac',
    description='Global and Robust Point Cloud Registration',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    ext_modules=ext_modules,
    install_requires=['pybind11>=2.2', 'numpy',
                      'open3d-python', 'cvxopt',
                      'picos', 'scipy'],
    cmdclass={'build_ext': BuildExt},
    zip_safe=False,
)
